@@ -62,7 +62,8 @@ class Model(object):
     @classmethod
     def log(cls, artifact_path, flavor, registered_model_name=None, **kwargs):
         """
-        Log model using supplied flavor module.
+        Log model using supplied flavor module. If no run is active, this method will create a new
+        active run.
 
         :param artifact_path: Run relative path identifying the model.
         :param flavor: Flavor module to save the model with. The module must have
@@ -130,12 +131,20 @@ class FlavorBackend(object):
         """
         pass
 
+    def prepare_env(self, model_uri):
+        """
+        Performs any preparation necessary to predict or serve the model, for example
+        downloading dependencies or initializing a conda environment. After preparation,
+        calling predict or serve should be fast.
+        """
+        pass
+
     @abstractmethod
     def can_score_model(self):
         """
         Check whether this flavor backend can be deployed in the current environment.
 
-        :return: True if this flavor backend can be applied int he current environment.
+        :return: True if this flavor backend can be applied in the current environment.
         """
         pass
 
