@@ -24,7 +24,8 @@ from __future__ import absolute_import
 import os
 import yaml
 import logging
-import posixpath
+import re
+import traceback
 
 import mlflow
 from mlflow import pyfunc, mleap
@@ -34,9 +35,12 @@ from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
 from mlflow.tracking.artifact_utils import _download_artifact_from_uri
 from mlflow.utils.environment import _mlflow_conda_env
 from mlflow.store.artifact.runs_artifact_repo import RunsArtifactRepository
+from mlflow.store.artifact.models_artifact_repo import ModelsArtifactRepository
 from mlflow.utils.file_utils import TempDir
-from mlflow.utils.uri import is_local_uri
+from mlflow.utils.uri import is_local_uri, append_to_uri_path
 from mlflow.utils.model_utils import _get_flavor_configuration_from_uri
+from mlflow.utils import experimental
+
 
 FLAVOR_NAME = "spark"
 
@@ -47,16 +51,28 @@ _SPARK_MODEL_PATH_SUB = "sparkml"
 _logger = logging.getLogger(__name__)
 
 
+def _format_exception(ex):
+    return "".join(traceback.format_exception(type(ex), ex, ex.__traceback__))
+
+
 def get_default_conda_env():
     """
     :return: The default Conda environment for MLflow Models produced by calls to
-             :func:`save_model()` and :func:`log_model()`.
+             :func:`save_model()` and :func:`log_model()`. This Conda environment
+             contains the current version of PySpark that is installed on the caller's
+             system. ``dev`` versions of PySpark are replaced with stable versions in
+             the resulting Conda environment (e.g., if you are running PySpark version
+             ``2.4.5.dev0``, invoking this method produces a Conda environment with a
+             dependency on PySpark version ``2.4.5``).
     """
     import pyspark
+    # Strip the suffix from `dev` versions of PySpark, which are not
+    # available for installation from Anaconda or PyPI
+    pyspark_version = re.sub(r"(\.?)dev.*", "", pyspark.__version__)
 
     return _mlflow_conda_env(
         additional_conda_deps=[
-            "pyspark={}".format(pyspark.__version__),
+            "pyspark={}".format(pyspark_version),
         ],
         additional_pip_deps=None,
         additional_conda_channels=None)
@@ -68,11 +84,13 @@ def log_model(spark_model, artifact_path, conda_env=None, dfs_tmpdir=None,
     Log a Spark MLlib model as an MLflow artifact for the current run. This uses the
     MLlib persistence format and produces an MLflow Model with the Spark flavor.
 
+    Note: If no run is active, it will instantiate a run to obtain a run_id.
+
     :param spark_model: Spark model to be saved - MLFlow can only save descendants of
                         pyspark.ml.Model which implement MLReadable and MLWritable.
     :param artifact_path: Run relative artifact path.
     :param conda_env: Either a dictionary representation of a Conda environment or the path to a
-                      Conda environment yaml file. If provided, this decribes the environment
+                      Conda environment yaml file. If provided, this decsribes the environment
                       this model should be run in. At minimum, it should specify the dependencies
                       contained in :func:`get_default_conda_env()`. If `None`, the default
                       :func:`get_default_conda_env()` environment is added to the model.
@@ -317,7 +335,7 @@ def save_model(spark_model, path, mlflow_model=Model(), conda_env=None,
     :param path: Local path where the model is to be saved.
     :param mlflow_model: MLflow model config this flavor is being added to.
     :param conda_env: Either a dictionary representation of a Conda environment or the path to a
-                      Conda environment yaml file. If provided, this decribes the environment
+                      Conda environment yaml file. If provided, this decsribes the environment
                       this model should be run in. At minimum, it should specify the dependencies
                       contained in :func:`get_default_conda_env()`. If `None`, the default
                       :func:`get_default_conda_env()` environment is added to the model.
@@ -388,9 +406,11 @@ def load_model(model_uri, dfs_tmpdir=None):
                       - ``relative/path/to/local/model``
                       - ``s3://my_bucket/path/to/model``
                       - ``runs:/<mlflow_run_id>/run-relative/path/to/model``
+                      - ``models:/<model_name>/<model_version>``
+                      - ``models:/<model_name>/<stage>``
 
                       For more information about supported URI schemes, see
-                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/tracking.html#
+                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/concepts.html#
                       artifact-locations>`_.
     :param dfs_tmpdir: Temporary directory path on Distributed (Hadoop) File System (DFS) or local
                        filesystem if running in local mode. The model is loaded from this
@@ -412,8 +432,12 @@ def load_model(model_uri, dfs_tmpdir=None):
         runs_uri = model_uri
         model_uri = RunsArtifactRepository.get_underlying_uri(model_uri)
         _logger.info("'%s' resolved as '%s'", runs_uri, model_uri)
+    elif ModelsArtifactRepository.is_models_uri(model_uri):
+        runs_uri = model_uri
+        model_uri = ModelsArtifactRepository.get_underlying_uri(model_uri)
+        _logger.info("'%s' resolved as '%s'", runs_uri, model_uri)
     flavor_conf = _get_flavor_configuration_from_uri(model_uri, FLAVOR_NAME)
-    model_uri = posixpath.join(model_uri, flavor_conf["model_data"])
+    model_uri = append_to_uri_path(model_uri, flavor_conf["model_data"])
     return _load_model(model_uri=model_uri, dfs_tmpdir=dfs_tmpdir)
 
 
@@ -455,3 +479,52 @@ class _PyFuncModelWrapper(object):
         spark_df = self.spark.createDataFrame(pandas_df)
         return [x.prediction for x in
                 self.spark_model.transform(spark_df).select("prediction").collect()]
+
+
+@experimental
+def autolog():
+    """
+    Enables automatic logging of Spark datasource paths, versions (if applicable), and formats
+    when they are read. This method is not threadsafe and assumes a
+    `SparkSession
+    <https://spark.apache.org/docs/latest/api/python/pyspark.sql.html#pyspark.sql.SparkSession>`_
+    already exists with the
+    `mlflow-spark JAR
+    <http://mlflow.org/docs/latest/tracking.html#automatic-logging-from-spark-experimental>`_
+    attached. It should be called on the Spark driver, not on the executors (i.e. do not call
+    this method within a function parallelized by Spark). This API requires Spark 3.0 or above.
+
+    Datasource information is logged under the current active MLflow run. If no active run
+    exists, datasource information is cached in memory & logged to the next-created active run
+    (but not to successive runs). Note that autologging of Spark ML (MLlib) models is not currently
+    supported via this API. Datasource-autologging is best-effort, meaning that if Spark is under
+    heavy load or MLflow logging fails for any reason (e.g. if the MLflow server is unavailable),
+    logging may be dropped.
+
+    For any unexpected issues with autologging, check Spark driver and executor logs in addition
+    to stderr & stdout generated from your MLflow code - datasource information is pulled from
+    Spark, so logs relevant to debugging may show up amongst the Spark logs.
+
+    >>> import mlflow.spark
+    >>> from pyspark.sql import SparkSession
+    >>> # Create and persist some dummy data
+    >>> spark = SparkSession.builder\
+    >>>   .config("spark.jars.packages", "org.mlflow.mlflow-spark").getOrCreate()
+    >>> df = spark.createDataFrame([
+    ...   (4, "spark i j k"),
+    ...   (5, "l m n"),
+    ...   (6, "spark hadoop spark"),
+    ...   (7, "apache hadoop")], ["id", "text"])
+    >>> import tempfile
+    >>> tempdir = tempfile.mkdtemp()
+    >>> df.write.format("csv").save(tempdir)
+    >>> # Enable Spark datasource autologging.
+    >>> mlflow.spark.autolog()
+    >>> loaded_df = spark.read.format("csv").load(tempdir)
+    >>> # Call collect() to trigger a read of the Spark datasource. Datasource info
+    >>> # (path and format)is automatically logged to an MLflow run.
+    >>> loaded_df.collect()
+    >>> shutil.rmtree(tempdir) # clean up tempdir
+    """
+    from mlflow import _spark_autologging
+    _spark_autologging.autolog()
