@@ -1,10 +1,8 @@
 import logging
 import os
-import posixpath
 import sys
 
 import uuid
-import six
 
 from mlflow.entities import Experiment, Metric, Param, Run, RunData, RunInfo, RunStatus, RunTag, \
     ViewType, SourceType, ExperimentTag
@@ -24,6 +22,8 @@ from mlflow.utils.file_utils import (is_directory, list_subdirs, mkdir, exists,
                                      write_to, append_to, make_containing_dirs, mv, get_parent_dir,
                                      list_all, local_file_uri_to_path, path_to_local_file_uri)
 from mlflow.utils.search_utils import SearchUtils
+from mlflow.utils.string_utils import is_string_type
+from mlflow.utils.uri import append_to_uri_path
 
 _TRACKING_DIR_ENV_VAR = "MLFLOW_TRACKING_DIR"
 
@@ -167,10 +167,10 @@ class FileStore(AbstractStore):
 
     def _get_artifact_dir(self, experiment_id, run_uuid):
         _validate_run_id(run_uuid)
-        artifacts_dir = posixpath.join(self.get_experiment(experiment_id).artifact_location,
-                                       run_uuid,
-                                       FileStore.ARTIFACTS_FOLDER_NAME)
-        return artifacts_dir
+        return append_to_uri_path(
+            self.get_experiment(experiment_id).artifact_location,
+            run_uuid,
+            FileStore.ARTIFACTS_FOLDER_NAME)
 
     def _get_active_experiments(self, full_path=False):
         exp_list = list_subdirs(self.root_directory, full_path)
@@ -200,7 +200,8 @@ class FileStore(AbstractStore):
         return experiments
 
     def _create_experiment_with_id(self, name, experiment_id, artifact_uri):
-        artifact_uri = artifact_uri or posixpath.join(self.artifact_root_uri, str(experiment_id))
+        artifact_uri = artifact_uri or append_to_uri_path(
+            self.artifact_root_uri, str(experiment_id))
         self._check_root_dir()
         meta_dir = mkdir(self.root_directory, str(experiment_id))
         experiment = Experiment(experiment_id, name, artifact_uri, LifecycleStage.ACTIVE)
@@ -346,7 +347,7 @@ class FileStore(AbstractStore):
 
     def update_run_info(self, run_id, run_status, end_time):
         _validate_run_id(run_id)
-        run_info = self.get_run(run_id).info
+        run_info = self._get_run_info(run_id)
         check_run_is_active(run_info)
         new_info = run_info._copy_with_overrides(run_status, end_time)
         self._overwrite_run_info(new_info)
@@ -396,9 +397,12 @@ class FileStore(AbstractStore):
         if run_info is None:
             raise MlflowException("Run '%s' metadata is in invalid state." % run_id,
                                   databricks_pb2.INVALID_STATE)
-        metrics = self.get_all_metrics(run_id)
-        params = self.get_all_params(run_id)
-        tags = self.get_all_tags(run_id)
+        return self._get_run_from_info(run_info)
+
+    def _get_run_from_info(self, run_info):
+        metrics = self._get_all_metrics(run_info)
+        params = self._get_all_params(run_info)
+        tags = self._get_all_tags(run_info)
         return Run(run_info, RunData(metrics, params, tags))
 
     def _get_run_info(self, run_uuid):
@@ -409,24 +413,19 @@ class FileStore(AbstractStore):
         if run_dir is None:
             raise MlflowException("Run '%s' not found" % run_uuid,
                                   databricks_pb2.RESOURCE_DOES_NOT_EXIST)
-
-        meta = read_yaml(run_dir, FileStore.META_DATA_FILE_NAME)
-        run_info = _read_persisted_run_info_dict(meta)
+        run_info = self._get_run_info_from_dir(run_dir)
         if run_info.experiment_id != exp_id:
-            logging.warning("Wrong experiment ID (%s) recorded for run '%s'. It should be %s. "
-                            "Run will be ignored.", str(run_info.experiment_id),
-                            str(run_info.run_id), str(exp_id), exc_info=True)
-            return None
-        return run_info
-
-    def _get_run_files(self, run_uuid, resource_type):
-        _validate_run_id(run_uuid)
-        run_info = self._get_run_info(run_uuid)
-        if run_info is None:
             raise MlflowException("Run '%s' metadata is in invalid state." % run_uuid,
                                   databricks_pb2.INVALID_STATE)
+        return run_info
 
-        _, run_dir = self._find_run_root(run_uuid)
+    def _get_run_info_from_dir(self, run_dir):
+        meta = read_yaml(run_dir, FileStore.META_DATA_FILE_NAME)
+        run_info = _read_persisted_run_info_dict(meta)
+        return run_info
+
+    def _get_run_files(self, run_info, resource_type):
+        run_dir = self._get_run_dir(run_info.experiment_id, run_info.run_id)
         # run_dir exists since run validity has been confirmed above.
         if resource_type == "metric":
             subfolder_name = FileStore.METRICS_FOLDER_NAME
@@ -476,7 +475,11 @@ class FileStore(AbstractStore):
 
     def get_all_metrics(self, run_uuid):
         _validate_run_id(run_uuid)
-        parent_path, metric_files = self._get_run_files(run_uuid, "metric")
+        run_info = self._get_run_info(run_uuid)
+        return self._get_all_metrics(run_info)
+
+    def _get_all_metrics(self, run_info):
+        parent_path, metric_files = self._get_run_files(run_info, "metric")
         metrics = []
         for metric_file in metric_files:
             metrics.append(self._get_metric_from_file(parent_path, metric_file))
@@ -497,8 +500,13 @@ class FileStore(AbstractStore):
     def get_metric_history(self, run_id, metric_key):
         _validate_run_id(run_id)
         _validate_metric_name(metric_key)
-        parent_path, metric_files = self._get_run_files(run_id, "metric")
+        run_info = self._get_run_info(run_id)
+        return self._get_metric_history(run_info, metric_key)
+
+    def _get_metric_history(self, run_info, metric_key):
+        parent_path, metric_files = self._get_run_files(run_info, "metric")
         if metric_key not in metric_files:
+            run_id = run_info.run_id
             raise MlflowException("Metric '%s' not found under run '%s'" % (metric_key, run_id),
                                   databricks_pb2.RESOURCE_DOES_NOT_EXIST)
         return [FileStore._get_metric_from_line(metric_key, line)
@@ -517,7 +525,12 @@ class FileStore(AbstractStore):
         return Param(param_name, value)
 
     def get_all_params(self, run_uuid):
-        parent_path, param_files = self._get_run_files(run_uuid, "param")
+        _validate_run_id(run_uuid)
+        run_info = self._get_run_info(run_uuid)
+        return self._get_all_params(run_info)
+
+    def _get_all_params(self, run_info):
+        parent_path, param_files = self._get_run_files(run_info, "param")
         params = []
         for param_file in param_files:
             params.append(self._get_param_from_file(parent_path, param_file))
@@ -543,7 +556,12 @@ class FileStore(AbstractStore):
         return RunTag(tag_name, tag_data)
 
     def get_all_tags(self, run_uuid):
-        parent_path, tag_files = self._get_run_files(run_uuid, "tag")
+        _validate_run_id(run_uuid)
+        run_info = self._get_run_info(run_uuid)
+        return self._get_all_tags(run_info)
+
+    def _get_all_tags(self, run_info):
+        parent_path, tag_files = self._get_run_files(run_info, "tag")
         tags = []
         for tag_file in tag_files:
             tags.append(self._get_tag_from_file(parent_path, tag_file))
@@ -554,23 +572,28 @@ class FileStore(AbstractStore):
         if not self._has_experiment(experiment_id):
             return []
         experiment_dir = self._get_experiment_path(experiment_id, assert_exists=True)
-        run_uuids = list_all(experiment_dir,
-                             filter_func=lambda x:
-                             all([os.path.basename(os.path.normpath(x)) != reservedFolderName
-                                  for reservedFolderName in
-                                  FileStore.RESERVED_EXPERIMENT_FOLDERS]) and os.path.isdir(x),
-                             full_path=False)
+        run_dirs = list_all(experiment_dir,
+                            filter_func=lambda x:
+                            all([os.path.basename(os.path.normpath(x)) != reservedFolderName
+                                 for reservedFolderName in
+                                 FileStore.RESERVED_EXPERIMENT_FOLDERS]) and os.path.isdir(x),
+                            full_path=True)
         run_infos = []
-        for r_id in run_uuids:
+        for r_dir in run_dirs:
             try:
                 # trap and warn known issues, will raise unexpected exceptions to caller
-                run_info = self._get_run_info(r_id)
-                if run_info is None:
+                run_info = self._get_run_info_from_dir(r_dir)
+                if run_info.experiment_id != experiment_id:
+                    logging.warning("Wrong experiment ID (%s) recorded for run '%s'. "
+                                    "It should be %s. Run will be ignored.",
+                                    str(run_info.experiment_id), str(run_info.run_id),
+                                    str(experiment_id), exc_info=True)
                     continue
                 if LifecycleStage.matches_view_type(view_type, run_info.lifecycle_stage):
                     run_infos.append(run_info)
             except MissingConfigException as rnfe:
                 # trap malformed run exception and log warning
+                r_id = os.path.basename(r_dir)
                 logging.warning("Malformed run '%s'. Detailed error %s", r_id, str(rnfe),
                                 exc_info=True)
         return run_infos
@@ -585,7 +608,7 @@ class FileStore(AbstractStore):
         runs = []
         for experiment_id in experiment_ids:
             run_infos = self._list_run_infos(experiment_id, run_view_type)
-            runs.extend(self.get_run(r.run_id) for r in run_infos)
+            runs.extend(self._get_run_from_info(r) for r in run_infos)
         filtered = SearchUtils.filter(runs, filter_string)
         sorted_runs = SearchUtils.sort(filtered, order_by)
         runs, next_page_token = SearchUtils.paginate(sorted_runs, page_token, max_results)
@@ -594,16 +617,19 @@ class FileStore(AbstractStore):
     def log_metric(self, run_id, metric):
         _validate_run_id(run_id)
         _validate_metric_name(metric.key)
-        run = self.get_run(run_id)
-        check_run_is_active(run.info)
-        metric_path = self._get_metric_path(run.info.experiment_id, run_id, metric.key)
+        run_info = self._get_run_info(run_id)
+        check_run_is_active(run_info)
+        self._log_run_metric(run_info, metric)
+
+    def _log_run_metric(self, run_info, metric):
+        metric_path = self._get_metric_path(run_info.experiment_id, run_info.run_id, metric.key)
         make_containing_dirs(metric_path)
         append_to(metric_path, "%s %s %s\n" % (metric.timestamp, metric.value, metric.step))
 
     def _writeable_value(self, tag_value):
         if tag_value is None:
             return ""
-        elif isinstance(tag_value, six.string_types):
+        elif is_string_type(tag_value):
             return tag_value
         else:
             return "%s" % tag_value
@@ -611,9 +637,12 @@ class FileStore(AbstractStore):
     def log_param(self, run_id, param):
         _validate_run_id(run_id)
         _validate_param_name(param.key)
-        run = self.get_run(run_id)
-        check_run_is_active(run.info)
-        param_path = self._get_param_path(run.info.experiment_id, run_id, param.key)
+        run_info = self._get_run_info(run_id)
+        check_run_is_active(run_info)
+        self._log_run_param(run_info, param)
+
+    def _log_run_param(self, run_info, param):
+        param_path = self._get_param_path(run_info.experiment_id, run_info.run_id, param.key)
         make_containing_dirs(param_path)
         write_to(param_path, self._writeable_value(param.value))
 
@@ -638,9 +667,12 @@ class FileStore(AbstractStore):
     def set_tag(self, run_id, tag):
         _validate_run_id(run_id)
         _validate_tag_name(tag.key)
-        run = self.get_run(run_id)
-        check_run_is_active(run.info)
-        tag_path = self._get_tag_path(run.info.experiment_id, run_id, tag.key)
+        run_info = self._get_run_info(run_id)
+        check_run_is_active(run_info)
+        self._set_run_tag(run_info, tag)
+
+    def _set_run_tag(self, run_info, tag):
+        tag_path = self._get_tag_path(run_info.experiment_id, run_info.run_id, tag.key)
         make_containing_dirs(tag_path)
         # Don't add trailing newline
         write_to(tag_path, self._writeable_value(tag.value))
@@ -652,12 +684,12 @@ class FileStore(AbstractStore):
         :param key: Name of the tag
         """
         _validate_run_id(run_id)
-        run = self.get_run(run_id)
-        check_run_is_active(run.info)
-        if key not in run.data.tags.keys():
+        run_info = self._get_run_info(run_id)
+        check_run_is_active(run_info)
+        tag_path = self._get_tag_path(run_info.experiment_id, run_id, key)
+        if not exists(tag_path):
             raise MlflowException("No tag with name: {} in run with id {}".format(key, run_id),
                                   error_code=RESOURCE_DOES_NOT_EXIST)
-        tag_path = self._get_tag_path(run.info.experiment_id, run_id, key)
         os.remove(tag_path)
 
     def _overwrite_run_info(self, run_info):
@@ -669,14 +701,14 @@ class FileStore(AbstractStore):
         _validate_run_id(run_id)
         _validate_batch_log_data(metrics, params, tags)
         _validate_batch_log_limits(metrics, params, tags)
-        run = self.get_run(run_id)
-        check_run_is_active(run.info)
+        run_info = self._get_run_info(run_id)
+        check_run_is_active(run_info)
         try:
             for param in params:
-                self.log_param(run_id, param)
+                self._log_run_param(run_info, param)
             for metric in metrics:
-                self.log_metric(run_id, metric)
+                self._log_run_metric(run_info, metric)
             for tag in tags:
-                self.set_tag(run_id, tag)
+                self._set_run_tag(run_info, tag)
         except Exception as e:
             raise MlflowException(e, INTERNAL_ERROR)
