@@ -27,6 +27,7 @@ import mlflow
 import tensorflow
 import mlflow.keras
 from distutils.version import LooseVersion
+from contextlib import contextmanager
 from tensorflow.keras.callbacks import Callback, TensorBoard  # pylint: disable=import-error
 from mlflow import pyfunc
 from mlflow.exceptions import MlflowException
@@ -37,7 +38,7 @@ from mlflow.utils import keyword_only, experimental
 from mlflow.utils.environment import _mlflow_conda_env
 from mlflow.utils.file_utils import _copy_file_or_tree
 from mlflow.utils.model_utils import _get_flavor_configuration
-from mlflow.utils.autologging_utils import try_mlflow_log
+from mlflow.utils.autologging_utils import try_mlflow_log, log_fn_args_as_params
 from mlflow.entities import Metric
 
 
@@ -53,6 +54,9 @@ _metric_queue = []
 
 _thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)
 
+# For tracking if the run was started by autologging.
+_AUTOLOG_RUN_ID = None
+
 
 def get_default_conda_env():
     """
@@ -100,7 +104,7 @@ def log_model(tf_saved_model_dir, tf_meta_graph_tags, tf_signature_def_key, arti
                                  ``tf.saved_model.builder.SavedModelBuilder`` method.
     :param artifact_path: The run-relative path to which to log model artifacts.
     :param conda_env: Either a dictionary representation of a Conda environment or the path to a
-                      Conda environment yaml file. If provided, this decribes the environment
+                      Conda environment yaml file. If provided, this decsribes the environment
                       this model should be run in. At minimum, it should specify the dependencies
                       contained in :func:`get_default_conda_env()`. If ``None``, the default
                       :func:`get_default_conda_env()` environment is added to the model. The
@@ -149,7 +153,7 @@ def save_model(tf_saved_model_dir, tf_meta_graph_tags, tf_signature_def_key, pat
     :param path: Local path where the MLflow model is to be saved.
     :param mlflow_model: MLflow model configuration to which to add the ``tensorflow`` flavor.
     :param conda_env: Either a dictionary representation of a Conda environment or the path to a
-                      Conda environment yaml file. If provided, this decribes the environment
+                      Conda environment yaml file. If provided, this decsribes the environment
                       this model should be run in. At minimum, it should specify the dependencies
                       contained in :func:`get_default_conda_env()`. If ``None``, the default
                       :func:`get_default_conda_env()` environment is added to the model. The
@@ -227,9 +231,11 @@ def load_model(model_uri, tf_sess=None):
                       - ``relative/path/to/local/model``
                       - ``s3://my_bucket/path/to/model``
                       - ``runs:/<mlflow_run_id>/run-relative/path/to/model``
+                      - ``models:/<model_name>/<model_version>``
+                      - ``models:/<model_name>/<stage>``
 
                       For more information about supported URI schemes, see
-                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/tracking.html#
+                      `Referencing Artifacts <https://www.mlflow.org/docs/latest/concepts.html#
                       artifact-locations>`_.
 
 
@@ -451,8 +457,7 @@ class __MLflowTfKerasCallback(Callback):
         Records model structural information as params after training finishes.
     """
     def __init__(self):
-        if mlflow.active_run() is None:
-            mlflow.start_run()
+        pass
 
     def __enter__(self):
         pass
@@ -460,10 +465,7 @@ class __MLflowTfKerasCallback(Callback):
     def __exit__(self, exc_type, exc_val, exc_tb):
         pass
 
-    def on_epoch_end(self, epoch, logs=None):
-        pass
-
-    def on_train_end(self, logs=None):  # pylint: disable=unused-argument
+    def on_train_begin(self, logs=None):  # pylint: disable=unused-argument
         opt = self.model.optimizer
         if hasattr(opt, 'optimizer'):
             opt = opt.optimizer
@@ -475,21 +477,35 @@ class __MLflowTfKerasCallback(Callback):
             epsilon = opt._epsilon if type(opt._epsilon) is float \
                 else tensorflow.keras.backend.eval(opt._epsilon)
             try_mlflow_log(mlflow.log_param, 'epsilon', epsilon)
-        l = []
-        self.model.summary(print_fn=l.append)
-        summary = '\n'.join(l)
-        try_mlflow_log(mlflow.set_tag, 'summary', summary)
+
+        sum_list = []
+        self.model.summary(print_fn=sum_list.append)
+        summary = '\n'.join(sum_list)
+        try_mlflow_log(mlflow.set_tag, 'model_summary', summary)
+
+        tempdir = tempfile.mkdtemp()
+        try:
+            summary_file = os.path.join(tempdir, "model_summary.txt")
+            with open(summary_file, 'w') as f:
+                f.write(summary)
+            try_mlflow_log(mlflow.log_artifact, local_path=summary_file)
+        finally:
+            shutil.rmtree(tempdir)
+
+    def on_epoch_end(self, epoch, logs=None):
+        pass
+
+    def on_train_end(self, logs=None):  # pylint: disable=unused-argument
         try_mlflow_log(mlflow.keras.log_model, self.model, artifact_path='model')
 
 
 class __MLflowTfKeras2Callback(Callback):
     """
         Callback for auto-logging parameters and metrics in TensorFlow >= 2.0.0.
-        Records model structural information as params after training finishes.
+        Records model structural information as params when training starts.
     """
     def __init__(self):
-        if mlflow.active_run() is None:
-            mlflow.start_run()
+        pass
 
     def __enter__(self):
         pass
@@ -497,19 +513,30 @@ class __MLflowTfKeras2Callback(Callback):
     def __exit__(self, exc_type, exc_val, exc_tb):
         pass
 
+    def on_train_begin(self, logs=None):  # pylint: disable=unused-argument
+        config = self.model.optimizer.get_config()
+        for attribute in config:
+            try_mlflow_log(mlflow.log_param, "opt_" + attribute, config[attribute])
+
+        sum_list = []
+        self.model.summary(print_fn=sum_list.append)
+        summary = '\n'.join(sum_list)
+        try_mlflow_log(mlflow.set_tag, 'model_summary', summary)
+
+        tempdir = tempfile.mkdtemp()
+        try:
+            summary_file = os.path.join(tempdir, "model_summary.txt")
+            with open(summary_file, 'w') as f:
+                f.write(summary)
+            try_mlflow_log(mlflow.log_artifact, local_path=summary_file)
+        finally:
+            shutil.rmtree(tempdir)
+
     def on_epoch_end(self, epoch, logs=None):
         if (epoch-1) % _LOG_EVERY_N_STEPS == 0:
             try_mlflow_log(mlflow.log_metrics, logs, step=epoch)
 
     def on_train_end(self, logs=None):  # pylint: disable=unused-argument
-        opt = self.model.optimizer
-        config = opt.get_config()
-        for attribute in config:
-            try_mlflow_log(mlflow.log_param, "opt_" + attribute, config[attribute])
-        l = []
-        self.model.summary(print_fn=l.append)
-        summary = '\n'.join(l)
-        try_mlflow_log(mlflow.set_tag, 'summary', summary)
         try_mlflow_log(mlflow.keras.log_model, self.model, artifact_path='model')
 
 
@@ -558,8 +585,10 @@ def _log_event(event):
     """
     Extracts metric information from the event protobuf
     """
-    if mlflow.active_run() is None:
-        mlflow.start_run()
+    if not mlflow.active_run():
+        try_mlflow_log(mlflow.start_run)
+        global _AUTOLOG_RUN_ID
+        _AUTOLOG_RUN_ID = mlflow.active_run().info.run_id
     if event.WhichOneof('what') == 'summary':
         summary = event.summary
         for v in summary.value:
@@ -586,15 +615,15 @@ def _setup_callbacks(lst):
     tb = _get_tensorboard_callback(lst)
     if tb is None:
         log_dir = tempfile.mkdtemp()
-        l = lst + [TensorBoard(log_dir)]
+        out_list = lst + [TensorBoard(log_dir)]
     else:
         log_dir = tb.log_dir
-        l = lst
+        out_list = lst
     if LooseVersion(tensorflow.__version__) < LooseVersion('2.0.0'):
-        l += [__MLflowTfKerasCallback()]
+        out_list += [__MLflowTfKerasCallback()]
     else:
-        l += [__MLflowTfKeras2Callback()]
-    return l, log_dir
+        out_list += [__MLflowTfKeras2Callback()]
+    return out_list, log_dir
 
 
 @experimental
@@ -631,8 +660,47 @@ def autolog(every_n_iter=100):
                       "1.12 <= v <= 2.0.0 are supported.")
         return
 
+    @contextmanager
+    def _manage_active_run():
+        if not mlflow.active_run():
+            try_mlflow_log(mlflow.start_run)
+            global _AUTOLOG_RUN_ID
+            if mlflow.active_run() is not None:  # defensive check in case `mlflow.start_run` fails
+                _AUTOLOG_RUN_ID = mlflow.active_run().info.run_id
+        yield mlflow.active_run()
+        if mlflow.active_run() is not None and mlflow.active_run().info.run_id == _AUTOLOG_RUN_ID:
+            try_mlflow_log(mlflow.end_run)
+
+    @gorilla.patch(tensorflow.estimator.Estimator)
+    def train(self, *args, **kwargs):
+        with _manage_active_run():
+            original = gorilla.get_original_attribute(tensorflow.estimator.Estimator, 'train')
+
+            # Checking step and max_step parameters for logging
+            if len(args) >= 3:
+                try_mlflow_log(mlflow.log_param, 'steps', args[2])
+                if len(args) >= 4:
+                    try_mlflow_log(mlflow.log_param, 'max_steps', args[3])
+            if 'steps' in kwargs:
+                try_mlflow_log(mlflow.log_param, 'steps', kwargs['steps'])
+            if 'max_steps' in kwargs:
+                try_mlflow_log(mlflow.log_param, 'max_steps', kwargs['max_steps'])
+
+            result = original(self, *args, **kwargs)
+
+            return result
+
     @gorilla.patch(tensorflow.estimator.Estimator)
     def export_saved_model(self, *args, **kwargs):
+        auto_end = False
+        if not mlflow.active_run():
+            global _AUTOLOG_RUN_ID
+            if _AUTOLOG_RUN_ID:
+                try_mlflow_log(mlflow.start_run, _AUTOLOG_RUN_ID)
+            else:
+                try_mlflow_log(mlflow.start_run)
+                auto_end = True
+
         original = gorilla.get_original_attribute(tensorflow.estimator.Estimator,
                                                   'export_saved_model')
         serialized = original(self, *args, **kwargs)
@@ -640,10 +708,22 @@ def autolog(every_n_iter=100):
                        tf_meta_graph_tags=[tag_constants.SERVING],
                        tf_signature_def_key='predict',
                        artifact_path='model')
+        if (mlflow.active_run() is not None and mlflow.active_run().info.run_id == _AUTOLOG_RUN_ID)\
+                or auto_end:
+            try_mlflow_log(mlflow.end_run)
         return serialized
 
     @gorilla.patch(tensorflow.estimator.Estimator)
     def export_savedmodel(self, *args, **kwargs):
+        auto_end = False
+        global _AUTOLOG_RUN_ID
+        if not mlflow.active_run():
+            if _AUTOLOG_RUN_ID:
+                try_mlflow_log(mlflow.start_run, _AUTOLOG_RUN_ID)
+            else:
+                try_mlflow_log(mlflow.start_run)
+                auto_end = True
+
         original = gorilla.get_original_attribute(tensorflow.estimator.Estimator,
                                                   'export_savedmodel')
         serialized = original(self, *args, **kwargs)
@@ -651,25 +731,118 @@ def autolog(every_n_iter=100):
                        tf_meta_graph_tags=[tag_constants.SERVING],
                        tf_signature_def_key='predict',
                        artifact_path='model')
+        if (mlflow.active_run() is not None and mlflow.active_run().info.run_id == _AUTOLOG_RUN_ID)\
+                or auto_end:
+            try_mlflow_log(mlflow.end_run)
         return serialized
 
+    def _early_stop_check(callbacks):
+        for callback in callbacks:
+            if isinstance(callback, tensorflow.keras.callbacks.EarlyStopping):
+                return callback
+        return None
+
+    def _log_early_stop_callback_params(callback):
+        if callback:
+            try:
+                earlystopping_params = {'monitor': callback.monitor,
+                                        'min_delta': callback.min_delta,
+                                        'patience': callback.patience,
+                                        'baseline': callback.baseline,
+                                        'restore_best_weights': callback.restore_best_weights}
+                try_mlflow_log(mlflow.log_params, earlystopping_params)
+            except Exception:  # pylint: disable=W0703
+                return
+
+    def _get_early_stop_callback_attrs(callback):
+        try:
+            return callback.stopped_epoch, callback.restore_best_weights, callback.patience
+        except Exception:  # pylint: disable=W0703
+            return None
+
+    def _log_early_stop_callback_metrics(callback, history):
+        if callback:
+            callback_attrs = _get_early_stop_callback_attrs(callback)
+            if callback_attrs is None:
+                return
+            stopped_epoch, restore_best_weights, patience = callback_attrs
+            try_mlflow_log(mlflow.log_metric, 'stopped_epoch', stopped_epoch)
+            # Weights are restored only if early stopping occurs
+            if stopped_epoch != 0 and restore_best_weights:
+                restored_epoch = stopped_epoch - max(1, patience)
+                try_mlflow_log(mlflow.log_metric, 'restored_epoch', restored_epoch)
+                restored_metrics = {key: history.history[key][restored_epoch]
+                                    for key in history.history.keys()}
+                # Metrics are logged as 'epoch_loss' and 'epoch_acc' in TF 1.X
+                if LooseVersion(tensorflow.__version__) < LooseVersion('2.0.0'):
+                    if 'loss' in restored_metrics:
+                        restored_metrics['epoch_loss'] = restored_metrics.pop('loss')
+                    if 'acc' in restored_metrics:
+                        restored_metrics['epoch_acc'] = restored_metrics.pop('acc')
+                # Checking that a metric history exists
+                metric_key = next(iter(history.history), None)
+                if metric_key is not None:
+                    last_epoch = len(history.history[metric_key])
+                    try_mlflow_log(mlflow.log_metrics, restored_metrics, step=last_epoch)
+
     @gorilla.patch(tensorflow.keras.Model)
     def fit(self, *args, **kwargs):
-        original = gorilla.get_original_attribute(tensorflow.keras.Model, 'fit')
-        # Checking if the 'callback' argument of fit() is set
-        if len(args) >= 6:
-            l = list(args)
-            l[5], log_dir = _setup_callbacks(l[5])
-            args = tuple(l)
-        elif 'callbacks' in kwargs:
-            kwargs['callbacks'], log_dir = _setup_callbacks(kwargs['callbacks'])
-        else:
-            kwargs['callbacks'], log_dir = _setup_callbacks([])
-        result = original(self, *args, **kwargs)
-        _flush_queue()
-        _log_artifacts_with_warning(local_dir=log_dir, artifact_path='tensorboard_logs')
-        shutil.rmtree(log_dir)
-        return result
+        with _manage_active_run():
+            original = gorilla.get_original_attribute(tensorflow.keras.Model, 'fit')
+
+            unlogged_params = ['self', 'x', 'y', 'callbacks', 'validation_data', 'verbose']
+
+            log_fn_args_as_params(original, args, kwargs, unlogged_params)
+            early_stop_callback = None
+
+            # Checking if the 'callback' argument of fit() is set
+            if len(args) >= 6:
+                tmp_list = list(args)
+                early_stop_callback = _early_stop_check(tmp_list[5])
+                tmp_list[5], log_dir = _setup_callbacks(tmp_list[5])
+                args = tuple(tmp_list)
+            elif 'callbacks' in kwargs:
+                early_stop_callback = _early_stop_check(kwargs['callbacks'])
+                kwargs['callbacks'], log_dir = _setup_callbacks(kwargs['callbacks'])
+            else:
+                kwargs['callbacks'], log_dir = _setup_callbacks([])
+
+            _log_early_stop_callback_params(early_stop_callback)
+
+            history = original(self, *args, **kwargs)
+
+            _log_early_stop_callback_metrics(early_stop_callback, history)
+
+            _flush_queue()
+            _log_artifacts_with_warning(local_dir=log_dir, artifact_path='tensorboard_logs')
+            shutil.rmtree(log_dir)
+
+            return history
+
+    @gorilla.patch(tensorflow.keras.Model)
+    def fit_generator(self, *args, **kwargs):
+        with _manage_active_run():
+            original = gorilla.get_original_attribute(tensorflow.keras.Model, 'fit_generator')
+
+            unlogged_params = ['self', 'generator', 'callbacks', 'validation_data', 'verbose']
+
+            log_fn_args_as_params(original, args, kwargs, unlogged_params)
+
+            # Checking if the 'callback' argument of fit() is set
+            if len(args) >= 5:
+                tmp_list = list(args)
+                tmp_list[4], log_dir = _setup_callbacks(tmp_list[4])
+                args = tuple(tmp_list)
+            elif 'callbacks' in kwargs:
+                kwargs['callbacks'], log_dir = _setup_callbacks(kwargs['callbacks'])
+            else:
+                kwargs['callbacks'], log_dir = _setup_callbacks([])
+            result = original(self, *args, **kwargs)
+            _flush_queue()
+            _log_artifacts_with_warning(local_dir=log_dir, artifact_path='tensorboard_logs')
+            shutil.rmtree(log_dir)
+
+            return result
 
     @gorilla.patch(EventFileWriter)
     def add_event(self, event):
@@ -688,7 +861,9 @@ def autolog(every_n_iter=100):
     patches = [
         gorilla.Patch(EventFileWriter, 'add_event', add_event, settings=settings),
         gorilla.Patch(EventFileWriterV2, 'add_event', add_event, settings=settings),
+        gorilla.Patch(tensorflow.estimator.Estimator, 'train', train, settings=settings),
         gorilla.Patch(tensorflow.keras.Model, 'fit', fit, settings=settings),
+        gorilla.Patch(tensorflow.keras.Model, 'fit_generator', fit_generator, settings=settings),
         gorilla.Patch(tensorflow.estimator.Estimator, 'export_saved_model',
                       export_saved_model, settings=settings),
         gorilla.Patch(tensorflow.estimator.Estimator, 'export_savedmodel',
